'use strict';

var _ast = require('../ast/ast.js');

var _ast2 = _interopRequireDefault(_ast);

var _makeAst = require('../make-ast.js');

var _makeAst2 = _interopRequireDefault(_makeAst);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var assertType = function assertType(schema, ft) {
  expect((0, _makeAst2.default)(schema)).toEqual(ft);
};

describe('makeAst', function () {
  describe('allOf', function () {
    it('works with primitive types', function () {
      var schema = {
        allOf: [{ type: 'string' }, { type: 'null' }]
      };

      var expected = _ast2.default.Intersection([_ast2.default.String, _ast2.default.Null]);
      assertType(schema, expected);
    });
  });
  describe('anyOf', function () {
    it('produces a union', function () {
      var schema = {
        anyOf: [{ type: 'string' }, { type: 'null' }]
      };

      var expected = _ast2.default.Union([_ast2.default.String, _ast2.default.Null]);
      assertType(schema, expected);
    });

    it('removes duplicates', function () {
      var schema = {
        anyOf: [{ type: 'string' }, { type: 'string' }, { type: 'null' }]
      };

      var expected = _ast2.default.Union([_ast2.default.String, _ast2.default.Null]);
      assertType(schema, expected);
    });
  });
  describe('enum', function () {
    it('produces a union of exact types', function () {
      var schema = {
        enum: [1, 'foo', { a: 1 }]
      };

      var expected = _ast2.default.Union([_ast2.default.Exact(1), _ast2.default.Exact('foo'), _ast2.default.Exact({ a: 1 })]);
      assertType(schema, expected);
    });
  });
  describe('items', function () {
    it('works with single-schema items', function () {
      var schema = {
        type: 'array',
        items: { type: 'string' }
      };

      var expected = _ast2.default.Array(_ast2.default.String);
      assertType(schema, expected);
    });

    it('works with an array of schemas as items', function () {
      var schema = {
        type: 'array',
        items: [{ type: 'string' }, {}]
      };

      var expected = _ast2.default.Tuple([_ast2.default.String, _ast2.default.Mixed]);
      assertType(schema, expected);
    });
  });
  it('oneOf', function () {});
  describe('objects', function () {
    it('works with properties and required', function () {
      var schema = {
        type: 'object',
        properties: {
          foo: { type: 'number' },
          bar: { type: 'boolean' }
        },
        required: ['foo']
      };

      var expected = _ast2.default.Record({
        foo: _ast2.default.Number,
        bar: _ast2.default.Optional(_ast2.default.Boolean)
      });

      assertType(schema, expected);
    });
  });
  describe('type', function () {
    it('works with string types', function () {
      assertType({ type: 'boolean' }, _ast2.default.Boolean);
      assertType({ type: 'null' }, _ast2.default.Null);
      assertType({ type: 'number' }, _ast2.default.Number);
      assertType({ type: 'integer' }, _ast2.default.Number);
      assertType({ type: 'string' }, _ast2.default.String);
    });

    it('reduces an array a single string to the single type', function () {
      var schema = {
        type: ['string']
      };

      var expected = _ast2.default.String;
      assertType(schema, expected);
    });

    it('works with an array of strings', function () {
      var schema = {
        type: ['string', 'boolean', 'boolean']
      };

      var expected = _ast2.default.Union([_ast2.default.String, _ast2.default.Boolean]);
      assertType(schema, expected);
    });

    it('works with an array of mixed strings and schemas', function () {
      var schema = {
        type: ['string', { type: 'boolean' }, {
          type: 'object',
          properties: {
            foo: { type: 'number' }
          }
        }]
      };

      var expected = _ast2.default.Union([_ast2.default.String, _ast2.default.Boolean, _ast2.default.Record({ foo: _ast2.default.Optional(_ast2.default.Number) })]);
      assertType(schema, expected);
    });
  });
});